/*******************************************************************************
*
*	TITLE:		control.c 
*
*	VERSION:	0.1                           
*
*	DATE:		01-29-2008
*
*	AUTHOR:		Team 1511
*
*	DESCRIPTION:	This file contains functions for controlling the arm and manipulator.
*******************************************************************************/

#include <stdio.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "control.h"
#include "iomap.h"
#include "motion.h"
#include "user_routines.h"

static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm);

static unsigned char Wrist_PID(unsigned int target);
static void Open_Manip(void);
static void Close_Manip(void);
static void Set_Pump_Precharge(int ticks);

unsigned int wrist_target;
static PIDData WristPID;
static int pump_precharge_ticks = 0;
static unsigned char arm_prev_up = 0;
static unsigned char manip_prev_open = 0;

void Initialize_Arm(void)
{
    wrist_target = anain_wrist_pot;
    Init_Wrist_PIDData_Struct(&WristPID, &pwm_wrist);
}

static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->tolerance = 10;
  control->kp_num = -3;
  control->kp_den = 1;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 127;
  control->i_out_max = 0;
  control->output_min = 25;
  control->output_max = 127;
}

void Arm_Control(void)
{
    if (oi_arm > 127 + ARM_CONTROL_DEADBAND)
        Raise_Arm();
    else if (oi_arm < 127 - ARM_CONTROL_DEADBAND)
        Lower_Arm();
    
    if (oi_gripper_open)
        Open_Manip();
    else if (oi_gripper_close)
        Close_Manip();

    if (oi_broken_sensors)
    {
        if (oi_wrist_up)
            pwm_wrist = 127 + WRIST_MANUAL_SPEED;
        else if (oi_wrist_down)
            pwm_wrist = 127 - WRIST_MANUAL_SPEED;
        else
            pwm_wrist = 127;
    }
    else
    {
        if (oi_wrist_up)
            wrist_target += WRIST_TARGET_STEP;
        else if (oi_wrist_down)
            wrist_target -= WRIST_TARGET_STEP;
        
        Wrist_PID(wrist_target);
    }

    if (pump_precharge_ticks)
        --pump_precharge_ticks;
}

int Raise_Arm(void)
{
    if (anain_wrist_pot < WRIST_POT_MIN_RAISE && !oi_broken_sensors)
    {
        if (wrist_target < WRIST_POT_MIN_RAISE)
            wrist_target = WRIST_POT_MIN_RAISE;
    }
    else
    {
        relay_arm_up = 1;
        relay_arm_down = 0;
        if (!arm_prev_up)
        {
            Set_Pump_Precharge(ARM_PUMP_TICKS);
            arm_prev_up = 1;
        }
    }
    return anain_arm_up;
}

int Lower_Arm(void)
{
    relay_arm_up = 0;
    relay_arm_down = 1;
    arm_prev_up = 0;
    return anain_arm_down;
}

static unsigned char Wrist_PID(unsigned int target)
{
    static unsigned int previous_position = 0;

    if ((anain_arm_up || relay_arm_up || anain_shoulder_pot < ARM_POT_MIN) && target < WRIST_POT_MIN_RAISE)
        target = WRIST_POT_MIN_RAISE;
    else if (target < WRIST_POT_MIN)
        target = WRIST_POT_MIN;
    else if (target > WRIST_POT_MAX)
        target = WRIST_POT_MAX;

    if (previous_position != target)
        Set_PIDData_Target(&WristPID, target);

    previous_position = target;
    return PID_Control(&WristPID, target);
}

static void Open_Manip(void)
{
    relay_gripper_open = 1;
    relay_gripper_close = 0;
    if (!manip_prev_open)
    {
        Set_Pump_Precharge(MANIP_PUMP_TICKS);
        manip_prev_open = 1;
    }
}

static void Close_Manip(void)
{
    relay_gripper_open = 0;
    relay_gripper_close = 1;
    if (manip_prev_open)
    {
        Set_Pump_Precharge(MANIP_PUMP_TICKS);
        manip_prev_open = 0;
    }
}

static void Set_Pump_Precharge(int ticks)
{
    if (pump_precharge_ticks < ticks)
        pump_precharge_ticks = ticks;
}

unsigned char Pump_Precharge(void)
{
    return (unsigned char)(pump_precharge_ticks > 0);
}
