/*******************************************************************************
*
*	TITLE:		motion.c 
*
*	VERSION:	1.0                           
*
*	DATE:		01-10-2008
*
*	AUTHOR:		Team 1511
*
*	DESCRIPTION:	This file contains functions for autonomously driving the wheels.
*******************************************************************************/

#include <stdio.h>
#include "motion.h"
#include "hybrid.h"
#include "iomap.h"
#include "encoder.h"
#include "adc.h"
#include "gyro.h"
#include "user_routines.h"
#include "control.h"

// Default pwm value limit for PID-based driving for Drive_Ticks()
#define DEFAULT_DRIVE_MAX 65

static void Init_Drive_PIDData_Struct(PIDData* control, unsigned char* pwm);

static PIDData pid_left;
static PIDData pid_right;

void Motion_Init(void)
{
  Init_Drive_PIDData_Struct(&pid_left, &pwm_left_drive);
  Init_Drive_PIDData_Struct(&pid_right, &pwm_right_drive);
}

void Init_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  control->target = 0;
  control->pwm = pwm;
  control->int_error = 0;
  control->prev_error = 0;
  control->tolerance = PID_NO_OSCILLATE;
  control->kp_num = 0;
  control->kp_den = 1;
  control->ki_num = 0;
  control->ki_den = 1;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 255;
  control->i_out_max = 255;
  control->output_min = 0;
  control->output_max = 255;
}

static void Init_Drive_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  Init_PIDData_Struct(control, pwm);
  control->kp_num = 1;
  control->kp_den = 16;
  control->ki_num = 5;
  control->ki_den = 1000;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 40;
  control->i_out_max = 30;
  control->output_min = 20;
  control->output_max = DEFAULT_DRIVE_MAX;
}



void Set_PIDData_Target(PIDData* control, long new_target)
{
  if ((control->tolerance == PID_NO_OSCILLATE) ||
      ((control->target > new_target) && (control->target - new_target > control->tolerance)) || 
      ((control->target < new_target) && (new_target - control->target > control->tolerance)))
    control->int_error = 0;
  control->target = new_target;
}


unsigned char PID_Control(PIDData* control, long current)
{
  if (! disabled_mode)
  {
    long error;
    long delta_error;
    int p_out, i_out, d_out, output;

    // Compute the error & delta error    
    error = control->target - current;
    delta_error = control->prev_error - error;

    // Save this error for next delta computation
    control->prev_error = error;

    // See if we've reached our destination; if so, stop.    
    if (control->tolerance == PID_NO_OSCILLATE) {
      if (((control->target > 0) && (error < 0)) || ((control->target < 0) && (error > 0))) {
        *(control->pwm) = 127;
        control->int_error = 0;
        return 1;
      }
    } else if (error < control->tolerance && error > -control->tolerance) {
      *(control->pwm) = 127;
      control->int_error = 0;
      return 1;
    }

    // Add up the integral component, and clamp it
    control->int_error += error;
    if (control->int_error > control->int_err_max)
      control->int_error = control->int_err_max;
    else if (control->int_error < -1 * control->int_err_max)
      control->int_error = -1 * control->int_err_max;

    // Compute p component, and clamp it    
    p_out = (int) (error * control->kp_num / control->kp_den);
    if (p_out > control->p_out_max)
      p_out = control->p_out_max;
    else if (p_out < -1 * control->p_out_max)
      p_out = -1 * control->p_out_max;

    // Compute i component, and clamp it
    i_out = (int) (control->int_error * control->ki_num / control->ki_den);
    if (i_out > control->i_out_max)
      i_out = control->i_out_max;
    else if (i_out < -1 * control->i_out_max)
      i_out = -1 * control->i_out_max;

    // Compute d component
    d_out = (int) (delta_error * control->kd_num / control->kd_den);
    
    // Sum PID components
    output = p_out + i_out + d_out;

    // Clamp to minimum value
    if (output > 0 && output < control->output_min)
      output = control->output_min;
    else if (output < 0 && output > -1 * control->output_min)
      output = -1 * control->output_min;

    if (output > control->output_max)
      output = control->output_max;
    else if (output < -1 * control->output_max)
      output = -1 * control->output_max;
    
    *(control->pwm) = (unsigned char) (output + 127);
     
    return 0;
  }
  return 0;
}

unsigned char Drive_Ticks(int heading, int integral_reset)
{
    long lcount, rcount;
    int lret, rret, angle, error=0, prop=0, deriv=0, lc=127, rc=127;
    static int integ=0, last_error=0;
    
    if (integral_reset)
	    {
            integ = 0;
            return 0;
        }

    lcount = Get_Left_Encoder_Count();
    rcount = Get_Right_Encoder_Count();

    lret = PID_Control(&pid_left, lcount);
    rret = PID_Control(&pid_right, rcount);
    
    if (!lret && !rret) 
    {
	    int lc = 127, rc = 127;

        angle = Get_Gyro_Angle();	//tenths of a degree
        error = (heading - angle);	//too far clockwise = negative num.
			
        prop = (int)error * (int)3 / (int)100;
        integ += error;
 	    deriv = (last_error - error) * (int)5 / (int)100;

        if (integ > 1000)	//limit integral windup
            integ = 1000;
        else if (integ < -1000)
             integ = -1000;

        if (prop > 20)
             prop = 20;
        else if (prop < -20)
             prop = -20;
    			
    	//40 times a second integral adds up
    	//error is in 0.1deg
    	//pwm range is 0-30
	    lc = (int)*(pid_left.pwm) + (prop) + (integ / (int)200);
	    rc = (int)*(pid_right.pwm) - (prop) - (integ/(int)200);
	
        if (lc > 255)
            lc = 255;
        else if (lc < 0)
            lc = 0;
    
        if (rc > 255)
            rc = 255;
        else if (rc < 0)
            rc = 0;

	*(pid_left.pwm) = (unsigned char)lc;
	*(pid_right.pwm) = (unsigned char)rc;
    }

  return lret && rret;
}

unsigned char Turn_Ticks(void)
{
  long left_ticks, right_ticks;
  int lret, rret;
  
  left_ticks = Get_Left_Encoder_Count();
  right_ticks = Get_Right_Encoder_Count();

  lret = PID_Control(&pid_left, pid_left.target > 0 ? -left_ticks : left_ticks);
  rret = PID_Control(&pid_right, pid_right.target > 0 ? right_ticks : -right_ticks);

  return lret || rret;
         
}
  
void Init_Drive_Turn(long ticks)
{
  Reset_Left_Encoder_Count();
  Reset_Right_Encoder_Count();
  
  Set_PIDData_Target(&pid_left, -ticks);
  Set_PIDData_Target(&pid_right, ticks);

  pid_left.output_max = DEFAULT_DRIVE_MAX;
  pid_right.output_max = DEFAULT_DRIVE_MAX;
}

void Init_Drive_Straight(long ticks)
{
  //Reset_Gyro_Angle();
  Init_Drive_Straight_Limited(ticks, DEFAULT_DRIVE_MAX);
}

void Init_Drive_Straight_Limited(long ticks, int max_power)
{
  Reset_Left_Encoder_Count();
  Reset_Right_Encoder_Count();
  
  Set_PIDData_Target(&pid_left, ticks);
  Set_PIDData_Target(&pid_right, ticks);

  pid_left.output_max = max_power;
  pid_right.output_max = max_power;
}



//work to be done:

int Turn_180(int lanenumber)    /* Turn 180 */
{
    int lane_distance;
    long angle=0;
    static int HEADING = -150;
    long R_Encoder = 0;


//    printf("We're in turn 180\r\n");
    angle=Get_Gyro_Angle();

    if (lanenumber == 0)
    {
        HEADING = -150;
        return 0;
    }
    else if (lanenumber == 1)
    {
        lane_distance = 42;
    }
    else if (lanenumber == 2)
    {
        lane_distance = 28;
    }
    else if (lanenumber == 3)
    {
        lane_distance = 14;
    }
    else if (lanenumber == 4)
    {
        lane_distance = 7;
    }


    R_Encoder = Get_Right_Encoder_Count();
    
    if (angle < -900)
        Drive_Angle(HEADING, 15, 0, DRIVE_FWDSPEED);
    else
        Drive_Angle(HEADING, 20, 0, DRIVE_FWDSPEED);

    if (R_Encoder > lane_distance*TICKS_PER_INCH )
    {    
        
        HEADING = HEADING -150;
        if (angle < -900)
            Drive_Angle(HEADING, 15, 1, DRIVE_FWDSPEED);
        else
            Drive_Angle(HEADING, 20, 1, DRIVE_FWDSPEED);

        Reset_Right_Encoder_Count(); 
        //last_error = 0;            
    }
 
    if (HEADING < -1800)
    {
        HEADING = -1800;
    }

    if (angle < -1750 && angle > -1850)
    {
        return 1;
    }
    return 0;
}
/* End Turn 180 */
                   
//call with fwdspeed=0 for a left turn, fwdspeed < 0 for a right turn
void Drive_Angle(int HEADING, int max_turn, int integral_reset, int fwdspeed)
{

	int angle, error=0, prop=0,deriv=0, lc=127, rc=127;
	static int integ=0, last_error=0;

    if (integral_reset)
        integ =0;

    if (fwdspeed <= 0)
        lc = rc = DRIVE_FWDSPEED;
    else
    {
        lc = fwdspeed - 4;
        rc= fwdspeed;

    }

    //printf("Hey we're in drive angle\r\n");
    angle = Get_Gyro_Angle();	//tenths of a degree
    error = (HEADING - angle);	//too far clockwise = negative num.
			
    //error of 15 degrees = 150
    //150 * 4 / 100 = pwm of 4
    prop = (int)error * (int)4 / (int)100;
    integ += error;
    deriv = (last_error - error) * (int)5 / (int)100;

    if (integ > 1000)	//limit integral windup
        integ = 1000;
    else if (integ < -1000)
         integ = -1000;

    if (prop > max_turn)
         prop = max_turn;
    else if (prop < -max_turn)
         prop = -max_turn;
    			
      //39 times a second integral adds up
      //error is in 0.1deg
      //1 degree off = error of 10
      //loop will have to execute 100 times before integral clamps
      //integral will clamp in 2.5 seconds for small errors
      //pwm range is -2 to +2 integral contribution
      

            printf("prop: %d error: %d\r\n", prop, error);
            printf("Gyro: %d\r\n", angle);
      //      printf("correction: %d", (prop + (integ/(int)500)));

    //base fwdspeed is already set
    lc += (prop) + (integ/(int)300); //- deriv);	//if error is negative, less left drive
    rc -= (prop) + (integ/(int)300); //- deriv);

    if (lc > 255)
        lc = 255;
    else if (lc < 127)
        lc = 127;
    
    if (rc > 255)
        rc = 255;
    else if (rc < 127)
        rc = 127;
       
    //fwdspeed = 0, left turn
    if (fwdspeed == 0)
        lc = 127;
    //fwdspeed < 0, right turn
    else if (fwdspeed < 0)
        rc = 127;
            
//         printf("Error: %d LeftCmd: %d RightCmd: %d\r\n", error, lc, rc);
    
    pwm_left_drive = (unsigned char)lc;
    pwm_right_drive = (unsigned char)rc;
//      printf("Leftpwm: %d Rightpwm: %d\r\n", (int)pwm_left_drive, (int)pwm_right_drive);
    
    
    //printf("Heading= %d Gyro= %d\r\n", HEADING, angle);
}




//comment out.
#if 0 
//variables that used for Changing Lanes
int Change_Lane_Mode =0;
int current_Lane = 1;// needs this variable linked to currnet info that tells which lane the robot belong
int Change_Lane_Done_Task =0;
int Change_Lane_SubMode = 0;
int current_Heading;
int Target_Lane_Holder;
int print_default_on_current_lane_holder =0;

//int Orgin_Angle; incomplete
/*
* wrote by jim bossert 
* Function name:  Change_Lane(int Target_Lane)
*
* Function's purpose is that a function that changes Lane by Robocoach commands and uses gyro to sneer 15 degree and 
* uses encoder to turn back to algin again after reached 208 inches
*
* passed int Target_Lane
*
* return int Change_Lane_Done_Task
*
*/
void Change_Lane(int Target_Lane)
{
	if (Change_Lane_Mode == 0)
	{
	if (print_default_on_current_lane_holder == 0)
	{
		current_Lane = default_Lane;
		print_default_on_current_lane_holder ++;
	}
	current_Heading  = Get_Gyro_Angle();
	Target_Lane_Holder = Target_Lane;
	//printf("In the loop Change_Lane_Mode %d \r", Change_Lane_Mode);
	Change_Lane_Mode++;
	}
	if (Change_Lane_Mode == 1)
		{
	//printf("In the loop Change_Lane_Mode %d \r", Change_Lane_Mode);	
			
			if (current_Lane == 1 && Target_Lane == 1)
			{
				Change_Lane_Mode =2;
				//printf(" the 1th chioces \r");
			}
			else if (current_Lane == 1 && Target_Lane == 2)
			{
				printf("Change_Lane_SubMode %d\r", Change_Lane_SubMode);
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Right();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_108_in();
				}
				else if (Change_Lane_SubMode == 4)  
				{
					return_to_Orgin_Heading();
				}
				//printf(" the 2th chioces \r");
			}
			else if (current_Lane == 1 && Target_Lane == 3)
			{
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Right();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_216_in();
				}
				else if (Change_Lane_SubMode == 4)
				{
					return_to_Orgin_Heading();
				}
				//printf(" the 3th chioces \r");
			}
			else if (current_Lane == 2 && Target_Lane == 1)
			{
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Left();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_108_in();
				}
				else if (Change_Lane_SubMode == 4)
				{
					return_to_Orgin_Heading();
				}
				//printf(" the 4th chioces \r");
			}
			else if (current_Lane == 2 && Target_Lane == 2)
			{
				Change_Lane_Mode =2;
				//printf(" the 5th chioces \r");
			}
			else if (current_Lane == 2 && Target_Lane == 3)
			{
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Right();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_108_in();
				}
				else if (Change_Lane_SubMode == 4)
				{
					return_to_Orgin_Heading();
				}
				//printf(" the 6th chioces \r");
			}
			else if (current_Lane == 3 && Target_Lane == 1)
			{
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Left();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_216_in();
				}
				else if (Change_Lane_SubMode == 4)
				{
					return_to_Orgin_Heading();
				}
				printf(" the 7th chioces \r");
			}
			else if (current_Lane == 3 && Target_Lane == 2)
			{
				if (Change_Lane_SubMode == 0 || Change_Lane_SubMode == 1)
				{
					Turn_15_Degree_Left();
				}
				else if (Change_Lane_SubMode == 2 || Change_Lane_SubMode == 3)
				{
					passes_108_in();
				}
				else if (Change_Lane_SubMode == 4)
				{
					return_to_Orgin_Heading();
				}
				//printf(" the 8th chioces \r");
			}
			else if (current_Lane == 3 && Target_Lane == 3)
			{
				Change_Lane_Mode =2;
				//printf(" the 9th chioces \r");
			}

		}
	else if (Change_Lane_Mode == 2)
	{
		current_Lane = 2;
		pwm_left_drive = 127;
		pwm_right_drive = 127;
		Change_Lane_Done_Task =1;
		Change_Lane_Mode = 0;
			//printf("In the loop Change_Lane_Mode %d \r", Change_Lane_Mode);
	}
	return;
}
/*
* wrote by jim bossert 
* Function name:  Turn_15_Degree Left()
*
* Function's purpose is that a robot turn left by 15 degree
*
* passed nothing
*
* return void
*
*/
void Turn_15_Degree_Left(void)
{
	if (Change_Lane_SubMode == 0)
	{
		Drive_Angle(current_Heading-500, 0, 0, DRIVE_FWDSPEED);
		Change_Lane_SubMode++;
		//printf("THE DRIVE ANGLE IS DIVINE AS IT IS FIRST TIMER as LEFT DIRECTION\r"); 
	}
	if (Change_Lane_SubMode == 1 && (Get_Gyro_Angle()-current_Heading)<-150)
	{
		Change_Lane_SubMode++;
	}
	else if (Change_Lane_SubMode == 1 && (Get_Gyro_Angle()-current_Heading)>-150)
	{
		Drive_Angle(current_Heading-500, 0, 0, DRIVE_FWDSPEED);
		//printf("THE DRIVE ANGLE IS CALLED ONCE AGAIN WELL, IT IS SO goood as drunk\r");
	}
	return;
}
/*
* wrote by jim bossert 
* Function name:  Turn_15_Degree_Right()
*
* Function's purpose is that a robot turn 15 degree right
*
* passed nothing
*
* return void
*
*/
void Turn_15_Degree_Right(void)
{
	if (Change_Lane_SubMode == 0)
	{
		Drive_Angle(current_Heading+500, 0, 0, DRIVE_FWDSPEED);  //drive angle will return 1 if it is done
		Change_Lane_SubMode++;
		//printf("THE DRIVE ANGLE IS DIVINE AS IT IS FIRST TIMER as RIGHT DIRECTION to home\r"); 
	}
	if (Change_Lane_SubMode == 1 && (Get_Gyro_Angle()-current_Heading)>150)
	{
		Change_Lane_SubMode++;
		Reset_Left_Encoder_Count();
    	Reset_Right_Encoder_Count();
	}
	else if (Change_Lane_SubMode == 1 && (Get_Gyro_Angle()-current_Heading)<150)
	{
		Drive_Angle(current_Heading+500, 0, 0, DRIVE_FWDSPEED);
		//printf("THE DRIVE ANGLE IS CALLED ONCE AGAIN WELL, IT IS SO goood as slam drunk\r");
	}
	return;
}
/*
* wrote by jim bossert 
* Function name:  passes_208_in()
*
* Function's purpose is that a robot passes 208 inches by encoder and straight drive
*
* passed nothing
*
* return void
*
*/
void passes_108_in(void)
{
	static int heading = 0;
	if (Change_Lane_SubMode == 2)
	{
		Init_Drive_Straight(72*TICKS_PER_INCH);
		heading = Get_Gyro_Angle();
		Change_Lane_SubMode++;
	}
	if (Drive_Ticks(heading, 0) == 1)
	{
		Change_Lane_SubMode++;
		
	}
	return;
}
/*
* wrote by jim bossert 
* Function name:  passes_416_in()
*
* Function's purpose is that a robot passes 416 inches by encoder and straight drive
*
* passed int nothing
*
* return void
*
*/
void passes_216_in(void)
{
	static int heading = 0;
	if (Change_Lane_SubMode ==2)
	{ 
		Init_Drive_Straight(190*TICKS_PER_INCH);
		heading = Get_Gyro_Angle();
		Change_Lane_SubMode++;
	}
	if (Drive_Ticks(heading, 0) == 1)
	{
		Change_Lane_SubMode++;
		
	}
	return;
}
/*
* wrote by jim bossert 
* Function name:  return_to_Orgin_Heading()
*
* Function's purpose is that a robot aglin again with its orgin heading
*
* passed nothing
*
* return void
*
*/
void return_to_Orgin_Heading(void)
{
	printf("OH HERE OH HERE TO ORIGIN HEADING FUNCTION WHICH RYMES WITH FUN\r");
	if (current_Heading < 0)
        Drive_Angle(current_Heading, 20, 0, 0);
    else
        Drive_Angle(current_Heading, 20, 0, -5);


	if (Get_Gyro_Angle() > -10 && Get_Gyro_Angle() < 10)
	{
		Change_Lane_Mode = 2;
		Reset_Left_Encoder_Count();
    	Reset_Right_Encoder_Count();
		current_Lane = Target_Lane_Holder;
		//printf("Here is finish with the return to origin\r");
	}
	
	return;
}

int Get_Task_Done(void)
{
	return Change_Lane_Done_Task;
}


//this brunch of local variables are only used by next two function below here
int Final_Left_Speed = 0;
int Final_Right_Speed = 0;
int Left_Adjust_rate = 0;
int Right_Adjust_rate = 0;

/*
* wrote by jim bossert 
* Function name:  Angle_Rate_Turn_TW()
*
* Function's purpose is that a robot turns based on the rate of change angle from gyro
*	not based on heading angle, to get more true smooth turn on two wheels.
*
* passed Left_Target_PWM, Right_Target_PWM, int rate_Angle
*
* return void
*
*/
void Angle_Rate_Turn_TW(int Left_Target_PWM, int Right_Target_PWM, int rate_Angle)
{
	//this logic statement is about finding if the robot turning right
	if (rate_Angle > 0)
		{
			if (Get_Gyro_Rate() > rate_Angle)
				{
					Right_Adjust_rate --;
					Check_Right_Adjust_rate(Right_Target_PWM);
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			else if (Get_Gyro_Rate() == rate_Angle)
				{
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			else if (Get_Gyro_Rate() < rate_Angle)
				{
					Right_Adjust_rate ++;
					Check_Right_Adjust_rate(Right_Target_PWM);
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			Final_Left_Speed = Left_Target_PWM;
		}
	//this logic statement is about finding if the robot turning left
	else if (rate_Angle < 0)
		{
			if (Get_Gyro_Rate() > rate_Angle)
				{
					Left_Adjust_rate --;
					Check_Left_Adjust_rate(Left_Target_PWM);
					Final_Left_Speed = Left_Adjust_rate + Left_Target_PWM;
				}
			else if (Get_Gyro_Rate() == rate_Angle)
				{
					Final_Left_Speed = Left_Adjust_rate + Right_Target_PWM;
				}
			else if (Get_Gyro_Rate() < rate_Angle)
				{
					Left_Adjust_rate ++;
					Check_Left_Adjust_rate(Left_Target_PWM);
					Final_Left_Speed = Left_Adjust_rate + Left_Target_PWM;
				}
			Final_Right_Speed = Right_Target_PWM;
		}

	return;	
	
}
/*
* wrote by jim bossert 
* Function name:  Angle_Rate_Turn_SW()
*
* Function's purpose is that a robot turns based on the rate of change angle from gyro
*	not based on heading angle, to get more true smooth turn on one wheels.
*
* passed Left_Target_PWM, Right_Target_PWM, int rate_Angle
*
* return void
*
*/
void Angle_Rate_Turn_SW(int Left_Target_PWM, int Right_Target_PWM, int rate_Angle)
{
	//this logic statement is about finding if the robot turning right
	if (rate_Angle > 0)
		{
			if (Get_Gyro_Rate() > rate_Angle)
				{
					Left_Adjust_rate --;
					Check_Left_Adjust_rate(Left_Target_PWM);
					Final_Left_Speed = Left_Adjust_rate + Left_Target_PWM;
				}
			else if (Get_Gyro_Rate() == rate_Angle)
				{
					Final_Left_Speed = Left_Adjust_rate + Left_Target_PWM;
				}
			else if (Get_Gyro_Rate() < rate_Angle)
				{
					Left_Adjust_rate ++;
					Check_Left_Adjust_rate(Left_Target_PWM);
					Final_Left_Speed = Left_Adjust_rate + Left_Target_PWM;
				}
			Final_Right_Speed = 127;
		}
	//this logic statement is about finding if the robot turning left
	else if (rate_Angle < 0)
		{
			if (Get_Gyro_Rate() > rate_Angle)
				{
					Right_Adjust_rate --;
					Check_Right_Adjust_rate(Right_Target_PWM);
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			else if (Get_Gyro_Rate() == rate_Angle)
				{
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			else if (Get_Gyro_Rate() < rate_Angle)
				{
					Right_Adjust_rate ++;
					Check_Right_Adjust_rate(Right_Target_PWM);
					Final_Right_Speed = Right_Adjust_rate + Right_Target_PWM;
				}
			Final_Left_Speed = 127;
		}
	pwm_left_drive = Final_Left_Speed;
	pwm_right_drive = Final_Right_Speed;

	return;	
}
/*
* wrote by jim bossert 
* Function name:  Check_Right_Adjust_rate()
*
* Function's purpose is check Right_Adjust_rate don't causes pwm values in -1 or 256
*
* passed Right_Target_PWM
*
* return void
*
*/
void Check_Right_Adjust_rate(int Right_Target_pwn)
{
	for (;(Right_Target_pwn + Right_Adjust_rate) > 255;)
	{
		Right_Adjust_rate --;
	}
	for (;(Right_Target_pwn + Right_Adjust_rate) < 0;)
	{
		Right_Adjust_rate ++;
	}
	
	return;
}
/*
* wrote by jim bossert 
* Function name:  Check_Left_Adjust_rate()
*
* Function's purpose is check Left_Adjust_rate don't causes pwm values in -1 or 256 or werid numbers
*
* passed Left_Target_PWM
*
* return void
*
*/
void Check_Left_Adjust_rate(int Left_Target_PWM)
{
	for (;(Left_Target_PWM + Left_Adjust_rate) > 255;)
	{
		Left_Adjust_rate --;
	}
	for (;(Left_Target_PWM + Left_Adjust_rate) < 0;)
	{
		Left_Adjust_rate ++;
	}
	
	return;
}
/*
* wrote by jim bossert 
* Function name:  reset_adjust_rates()
*
* Function's purpose is reset the adjust rates depended by drive function that based on rate of
*	angle 
*
* passed two bits to tell which rates to reset
*
* return void
*
*/
void reset_adjust_rates(char right_bit, char left_bit)
{
	if (right_bit == 1)
	Right_Adjust_rate = 0;
	if (left_bit == 1)
	Left_Adjust_rate = 0;

	return;
}

#endif
