/*******************************************************************************
* FILE NAME: user_routines.c <FRC VERSION>
*
* DESCRIPTION:
*  This file contains the default mappings of inputs  
*  (like switches, joysticks, and buttons) to outputs on the RC.  
*
* USAGE:
*  You can either modify this file to fit your needs, or remove it from your 
*  project and replace it with a modified copy. 
*
*******************************************************************************/

#include <stdio.h>

#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "user_routines.h"
#include "serial_ports.h"
#include "eeprom.h"
#include "tracking.h"
#include "camera.h"
#include "Visual_drive.h"
#include "control.h"
#include "motion.h"
#include "encoder.h"
#include "gyro.h"
#include "adc.h"

/* Prototypes */
static void Manual_Drive(void);
static void Calibrate_Gyro(void);
static void chicklet_analog_decode(void);
static unsigned char chicklet_analog_bit(unsigned char analogvalue, int bitpos);
static void Debug_Print(void);
static void Send_Feedback(void);
static void autonomous_decode(void);
static void position_decode(void);

// Drive code LUTs
static const rom unsigned char ten_pct[];
static const rom unsigned char drive_lut[];
static const rom unsigned char turbo_drive_lut[];
static const rom unsigned char slow_drive_lut[];

// Buttons mapped from Chicklet analog values
unsigned char oi_armpos_pickup = 0;
unsigned char oi_wrist_up = 0;
unsigned char oi_wrist_down = 0;


unsigned char automode = 0;
// at autoposition, 1 means left, 2 means middle, and 3 means right 
unsigned int autoposition = 0;



/*******************************************************************************
* FUNCTION NAME: User_Initialization
* PURPOSE:       This routine is called first (and only once) in the Main function.  
*                You may modify and add to this function.
* CALLED FROM:   main.c
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void User_Initialization (void)
{
  Set_Number_of_Analog_Channels(SIXTEEN_ANALOG);    /* DO NOT CHANGE! */

/* FIRST: Set up the I/O pins you want to use as digital INPUTS. */
  digital_io_01 = digital_io_02 = digital_io_03 = digital_io_04 = INPUT;
  digital_io_05 = digital_io_06 = digital_io_07 = digital_io_08 = INPUT;
  digital_io_09 = digital_io_10 = digital_io_11 = digital_io_12 = INPUT;
  digital_io_13 = digital_io_14 = digital_io_15 = digital_io_16 = INPUT;
  digital_io_18 = INPUT;  /* Used for pneumatic pressure switch. */
    /* 
     Note: digital_io_01 = digital_io_02 = ... digital_io_04 = INPUT; 
           is the same as the following:

           digital_io_01 = INPUT;
           digital_io_02 = INPUT;
           ...
           digital_io_04 = INPUT;
    */

/* SECOND: Set up the I/O pins you want to use as digital OUTPUTS. */
  digital_io_17 = OUTPUT;    /* Example - Not used in Default Code. */

/* THIRD: Initialize the values on the digital outputs. */
  rc_dig_out17 = 0;

/* FOURTH: Set your initial PWM values.  Neutral is 127. */
  pwm01 = pwm02 = pwm03 = pwm04 = pwm05 = pwm06 = pwm07 = pwm08 = 127;
  pwm09 = pwm10 = pwm11 = pwm12 = pwm13 = pwm14 = pwm15 = pwm16 = 127;

/* FIFTH: Set your PWM output types for PWM OUTPUTS 13-16.
  /*   Choose from these parameters for PWM 13-16 respectively:               */
  /*     IFI_PWM  - Standard IFI PWM output generated with Generate_Pwms(...) */
  /*     USER_CCP - User can use PWM pin as digital I/O or CCP pin.           */
  Setup_PWM_Output_Type(IFI_PWM,IFI_PWM,IFI_PWM,IFI_PWM);

  /* 
     Example: The following would generate a 40KHz PWM with a 50% duty cycle on the CCP2 pin:

         CCP2CON = 0x3C;
         PR2 = 0xF9;
         CCPR2L = 0x7F;
         T2CON = 0;
         T2CONbits.TMR2ON = 1;

         Setup_PWM_Output_Type(USER_CCP,IFI_PWM,IFI_PWM,IFI_PWM);
  */

  // Initialize drive control structure
  Motion_Init();

  Initialize_Encoders();

  Initialize_ADC();
  Initialize_Gyro();

  Initialize_Arm();
  
  // Set all PWMs to neutral; reset all state.
  Reset_All();

Init_Serial_Port_One();
Init_Serial_Port_Two();

#ifdef TERMINAL_SERIAL_PORT_1    
  stdout_serial_port = SERIAL_PORT_ONE;
#endif


  /* Add any other initialization code here. */
 
  Putdata(&txdata);             /* DO NOT CHANGE! */

  

  printf("IFI 2008 User Processor Initialized ...\r");  /* Optional - Print initialization message. */

  User_Proc_Is_Ready();         /* DO NOT CHANGE! - last line of User_Initialization */
}

/*
 * Resets all PWMs, Relays, and user_routines state.
 * Shared across initial initialization, enter, and exit of
 * autonomous.
 */
void Reset_All(void)
{
    // All PWMs to neutral
    pwm01 = pwm02 = pwm03 = pwm04 = pwm05 = pwm06 = pwm07 = pwm08 = 127;
    pwm09 = pwm10 = pwm11 = pwm12 = pwm13 = pwm14 = pwm15 = pwm16 = 127;
    
    // All Relays to neutral
    relay1_fwd = relay1_rev = relay2_fwd = relay2_rev = 0;
    relay3_fwd = relay3_rev = relay4_fwd = relay4_rev = 0;
    relay5_fwd = relay5_rev = relay6_fwd = relay6_rev = 0;
    relay7_fwd = relay7_rev = relay8_fwd = relay8_rev = 0;

    digout_coast = 0;
}


/*******************************************************************************
* FUNCTION NAME: Process_Data_From_Master_uP
* PURPOSE:       Executes every 26.2ms when it gets new data from the master 
*                microprocessor.
* CALLED FROM:   main.c
* ARGUMENTS:     none
* RETURNS:       void
*******************************************************************************/
void Process_Data_From_Master_uP(void)
{
  static unsigned char i;

  Getdata(&rxdata);   /* Get fresh data from the master microprocessor. */

	
	EEPROM_Write_Handler();
	Camera_Handler();
	

//	printf("Histogram %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d\r",
//		H_Packet_Data.histogram[0], H_Packet_Data.histogram[1], H_Packet_Data.histogram[2], H_Packet_Data.histogram[3],
//		H_Packet_Data.histogram[4], H_Packet_Data.histogram[5], H_Packet_Data.histogram[6], H_Packet_Data.histogram[7],
//		H_Packet_Data.histogram[8], H_Packet_Data.histogram[9], H_Packet_Data.histogram[10], H_Packet_Data.histogram[11],
//		H_Packet_Data.histogram[12], H_Packet_Data.histogram[13], H_Packet_Data.histogram[14], H_Packet_Data.histogram[15],
//		H_Packet_Data.histogram[16], H_Packet_Data.histogram[17], H_Packet_Data.histogram[18], H_Packet_Data.histogram[19],
//		H_Packet_Data.histogram[20], H_Packet_Data.histogram[21], H_Packet_Data.histogram[22], H_Packet_Data.histogram[23], 
//		H_Packet_Data.histogram[24], H_Packet_Data.histogram[25], H_Packet_Data.histogram[26], H_Packet_Data.histogram[27]);
	printf("Histogram %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d\r",
		H_Packet_Data.histogram[0], H_Packet_Data.histogram[1], H_Packet_Data.histogram[2], H_Packet_Data.histogram[3],
		H_Packet_Data.histogram[4], H_Packet_Data.histogram[5], H_Packet_Data.histogram[6], H_Packet_Data.histogram[7],
		H_Packet_Data.histogram[8], H_Packet_Data.histogram[9], H_Packet_Data.histogram[10], H_Packet_Data.histogram[11],
		H_Packet_Data.histogram[12], H_Packet_Data.histogram[13], H_Packet_Data.histogram[14], H_Packet_Data.histogram[15],
		H_Packet_Data.histogram[16], H_Packet_Data.histogram[17], H_Packet_Data.histogram[18], H_Packet_Data.histogram[19],
		H_Packet_Data.histogram[20], H_Packet_Data.histogram[21], H_Packet_Data.histogram[22], H_Packet_Data.histogram[23], 
		H_Packet_Data.histogram[24], H_Packet_Data.histogram[25], H_Packet_Data.histogram[26], H_Packet_Data.histogram[27]);
	

  /* Add your own code here. (a printf will not be displayed when connected to the breaker panel unless a Y cable is used) */

  //printf("Port1 Y %3d, X %3d, Fire %d, Top %d\r",pwm01,pwm05,p1_sw_trig,p1_sw_top);  /* printf EXAMPLE */

  Generate_Pwms(pwm13,pwm14,pwm15,pwm16);

   if(disabled_mode == 1)
{
	autonomous_decode();
	position_decode();
}
  chicklet_analog_decode();
  Arm_Control();

  Calibrate_Gyro();

  Manual_Drive();

  // On 2006 OI:
  // oi_debug_button1 == oi_arm_hi
  // oi_debug_button2 == oi_arm_mid
  if (oi_debug_switch) {
    /* Debug mode */
    Debug_Print();

  } else {
    /* Normal mode */

    if (oi_debug_button2) {
      pwm_left_drive = 187;
      pwm_right_drive = 187;
    }
  }

  Send_Feedback();



  Putdata(&txdata);             /* DO NOT CHANGE! */
}

static void autonomous_decode(void)
{
  int tens = 0;
  int ones = 0;

  if (oi_auto_tens_b0 > 200) tens |= 0x01;
  if (oi_auto_tens_b1 > 200) tens |= 0x02;
  if (oi_auto_tens_b2 > 200) tens |= 0x04;
  if (oi_auto_ones_b0 > 200) ones |= 0x01;
  if (oi_auto_ones_b0 > 200) ones |= 0x02;
  if (oi_auto_ones_b0 > 200) ones |= 0x04;
  
  automode = (unsigned char)((tens * 10) + ones);
}
/*
* wrote by jim bossert 
* Function name:  position_decode()
*
* Function's purpose is decode what oi tells rc which position the 
*	robot is in
* passed nothing
*
* return void
*
*/
static void position_decode(void)
{
	if (oi_starting_pos < 127)
		{
			autoposition = 1;
		}
	if(oi_starting_pos = 127)
		{
			autoposition = 2;
		}
	if (oi_starting_pos > 127)
		{
			autoposition = 3;
		}
}

/* chicklet_analog_decode converts 4 high bits of 
 * analog value (unsigned char) to button values by
 * setting the global oi_xxx variables.
 *
 * For reference on how buttons are mapped, see oi.xls
 */
static void chicklet_analog_decode(void)
{
    oi_wrist_up = chicklet_analog_bit(p4_aux, 5);
    oi_wrist_down = chicklet_analog_bit(p4_aux, 4);
    oi_armpos_pickup = chicklet_analog_bit(p4_wheel, 6);
}

static unsigned char chicklet_analog_bit(unsigned char analogvalue, int bitpos)
{
    // note: button pressed==0
    //       button not pressed==1
    int x = (int)analogvalue;
    x &= ((int)1) << bitpos;
    if (x == 0)
        return 1;
    else
        return 0;
}



static void Manual_Drive(void) {
    if (oi_slow_left && oi_turbo_left && oi_slow_right && oi_turbo_right)
    {
        pwm_right_drive = 255 - ten_pct[turbo_drive_lut[oi_left_drive]];
        pwm_left_drive = 255 - turbo_drive_lut[oi_right_drive];
    }
    else
    {
      if (oi_slow_left)
          pwm_left_drive = slow_drive_lut[oi_left_drive];
      else if (oi_turbo_left)
          pwm_left_drive = turbo_drive_lut[oi_left_drive];
      else
          pwm_left_drive = drive_lut[oi_left_drive];
      
      if (oi_slow_right)
          pwm_right_drive = ten_pct[slow_drive_lut[oi_right_drive]];
      else if (oi_turbo_right)
          pwm_right_drive = ten_pct[turbo_drive_lut[oi_right_drive]];
      else
          pwm_right_drive = ten_pct[drive_lut[oi_right_drive]];
    }
}


int getIR(void)
{
    int IR;

    if  (digin_ir1 == 1)
        IR = 1;
    else if  (digin_ir2 == 1)
        IR = 2;
    else if  (digin_ir3 == 1)
        IR = 3;
    else if  (digin_ir4 == 1)
        IR = 4;
    else 
        IR = 0;

    printf("IR %d\r\n", (int)IR);
    return IR;
}

static void Calibrate_Gyro(void) {
    static unsigned int j = 0;
    static unsigned int i = 0;
    int temp_gyro_rate;
    long temp_gyro_angle;
    int temp_gyro_bias;

  if (j == 300) {
    if (!i) {
      temp_gyro_angle = Get_Gyro_Angle();
      printf("Gyro Angle: %d\r\n", (int)temp_gyro_angle);
      i = 60;
    }
    --i;
    return;
  }

    j++;

    if(j == 10)
    {
        printf("Calculating Gyro Bias...\r\n");
    }

    if(j == 60)
    {
        // start a gyro bias calculation
        Start_Gyro_Bias_Calc();
    }

    if(j == 300)
    {
        // terminate the gyro bias calculation
        Stop_Gyro_Bias_Calc();

        // reset the gyro heading angle
        Reset_Gyro_Angle();

        printf("Gyro Bias Calculation Complete...\r\n");
    }
}

static void Send_Feedback(void)
{
  if (user_display_mode) {
    if (disabled_mode)
      User_Mode_byte = automode;
    else
      /* Send backup battery voltage */
      User_Mode_byte = backup_voltage*10;

  } else {
    oi_led_debug = oi_debug_switch;

  }
}

static void Debug_Print(void)
{
  static int pcount = 0;
  
  if (pcount++ < 30)
    return;
  
  pcount = 0;
  
  if (oi_debug_button1) {
    printf("Left  Enc:  %ld\n", Get_Left_Encoder_Count());
    printf("Right Enc:  %ld\n", Get_Right_Encoder_Count());
  }
  if (oi_debug_button2) {
    long x = Get_Gyro_Angle();
    printf("Gyro Angle:  %d.%d\n", (int)(x / 10), (int)(x % 10));
  }
}

 // Drive lookup tables
// For each function, xmin = -127 and xmax = 127

//(x^3) * (55/2048383) + 127
static const rom unsigned char drive_lut [] = {
    72, 73, 75, 76, 77, 78, 79, 81, 
    82, 83, 84, 85, 86, 87, 88, 89, 
    90, 91, 92, 93, 94, 95, 96, 97, 
    98, 99, 99, 100, 101, 102, 102, 103, 
    104, 105, 105, 106, 107, 107, 108, 109, 
    109, 110, 111, 111, 112, 112, 113, 113, 
    114, 114, 115, 115, 116, 116, 117, 117, 
    117, 118, 118, 119, 119, 119, 120, 120, 
    120, 121, 121, 121, 121, 122, 122, 122, 
    123, 123, 123, 123, 123, 124, 124, 124, 
    124, 124, 125, 125, 125, 125, 125, 125, 
    125, 126, 126, 126, 126, 126, 126, 126, 
    126, 126, 126, 126, 126, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 127, 127, 127, 127, 127, 127, 
    127, 127, 128, 128, 128, 128, 128, 128, 
    128, 128, 128, 128, 128, 128, 129, 129, 
    129, 129, 129, 129, 129, 130, 130, 130, 
    130, 130, 131, 131, 131, 131, 131, 132, 
    132, 132, 133, 133, 133, 133, 134, 134, 
    134, 135, 135, 135, 136, 136, 137, 137, 
    137, 138, 138, 139, 139, 140, 140, 141, 
    141, 142, 142, 143, 143, 144, 145, 145, 
    146, 147, 147, 148, 149, 149, 150, 151, 
    152, 152, 153, 154, 155, 155, 156, 157, 
    158, 159, 160, 161, 162, 163, 164, 165, 
    166, 167, 168, 169, 170, 171, 172, 173, 
    175, 176, 177, 178, 179, 181, 182, 183
};

static const rom unsigned char ten_pct [] = {
    13, 14, 15, 15, 16, 17, 18, 19, 
    20, 21, 22, 23, 24, 24, 25, 26, 
    27, 28, 29, 30, 31, 32, 33, 33, 
    34, 35, 36, 37, 38, 39, 40, 41, 
    42, 42, 43, 44, 45, 46, 47, 48, 
    49, 50, 51, 51, 52, 53, 54, 55, 
    56, 57, 58, 59, 60, 60, 61, 62, 
    63, 64, 65, 66, 67, 68, 69, 69, 
    70, 71, 72, 73, 74, 75, 76, 77, 
    78, 78, 79, 80, 81, 82, 83, 84, 
    85, 86, 87, 87, 88, 89, 90, 91, 
    92, 93, 94, 95, 96, 96, 97, 98, 
    99, 100, 101, 102, 103, 104, 105, 105, 
    106, 107, 108, 109, 110, 111, 112, 113, 
    114, 114, 115, 116, 117, 118, 119, 120, 
    121, 122, 123, 123, 124, 125, 126, 127, 
    128, 129, 130, 131, 132, 132, 133, 134, 
    135, 136, 137, 138, 139, 140, 141, 141, 
    142, 143, 144, 145, 146, 147, 148, 149, 
    150, 150, 151, 152, 153, 154, 155, 156, 
    157, 158, 159, 159, 160, 161, 162, 163, 
    164, 165, 166, 167, 168, 168, 169, 170, 
    171, 172, 173, 174, 175, 176, 177, 177, 
    178, 179, 180, 181, 182, 183, 184, 185, 
    186, 186, 187, 188, 189, 190, 191, 192, 
    193, 194, 195, 195, 196, 197, 198, 199, 
    200, 201, 202, 203, 204, 204, 205, 206, 
    207, 208, 209, 210, 211, 212, 213, 213, 
    214, 215, 216, 217, 218, 219, 220, 221, 
    222, 222, 223, 224, 225, 226, 227, 228, 
    229, 230, 231, 231, 232, 233, 234, 235, 
    236, 237, 238, 239, 240, 240, 241, 242
};

//(x^3) * (127/2048383) + 127
static const rom unsigned char turbo_drive_lut[] = {
    0x00, 0x03, 0x06, 0x09, 0x0c, 0x0e, 0x11, 0x14, 
    0x17, 0x19, 0x1c, 0x1e, 0x21, 0x23, 0x26, 0x28, 
    0x2a, 0x2c, 0x2f, 0x31, 0x33, 0x35, 0x37, 0x39, 
    0x3b, 0x3d, 0x3f, 0x41, 0x43, 0x45, 0x46, 0x48, 
    0x4a, 0x4c, 0x4d, 0x4f, 0x50, 0x52, 0x53, 0x55, 
    0x56, 0x58, 0x59, 0x5a, 0x5c, 0x5d, 0x5e, 0x5f, 
    0x60, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 
    0x69, 0x6a, 0x6b, 0x6c, 0x6c, 0x6d, 0x6e, 0x6f, 
    0x6f, 0x70, 0x71, 0x72, 0x72, 0x73, 0x74, 0x74, 
    0x75, 0x75, 0x76, 0x76, 0x77, 0x77, 0x78, 0x78, 
    0x79, 0x79, 0x79, 0x7a, 0x7a, 0x7a, 0x7b, 0x7b, 
    0x7b, 0x7c, 0x7c, 0x7c, 0x7c, 0x7d, 0x7d, 0x7d, 
    0x7d, 0x7d, 0x7d, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 
    0x7e, 0x7e, 0x7e, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x80, 0x80, 0x80, 0x80, 
    0x80, 0x80, 0x80, 0x80, 0x81, 0x81, 0x81, 0x81, 
    0x81, 0x81, 0x82, 0x82, 0x82, 0x82, 0x83, 0x83, 
    0x83, 0x84, 0x84, 0x84, 0x85, 0x85, 0x85, 0x86, 
    0x86, 0x87, 0x87, 0x88, 0x88, 0x89, 0x89, 0x8a, 
    0x8a, 0x8b, 0x8c, 0x8c, 0x8d, 0x8e, 0x8f, 0x8f, 
    0x90, 0x91, 0x92, 0x92, 0x93, 0x94, 0x95, 0x96, 
    0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9e, 0x9f, 
    0xa0, 0xa1, 0xa2, 0xa4, 0xa5, 0xa6, 0xa8, 0xa9, 
    0xab, 0xac, 0xae, 0xaf, 0xb1, 0xb2, 0xb4, 0xb6, 
    0xb8, 0xb9, 0xbb, 0xbd, 0xbf, 0xc1, 0xc3, 0xc5, 
    0xc7, 0xc9, 0xcb, 0xcd, 0xcf, 0xd2, 0xd4, 0xd6, 
    0xd8, 0xdb, 0xdd, 0xe0, 0xe2, 0xe5, 0xe7, 0xea, 
    0xed, 0xf0, 0xf2, 0xf5, 0xf8, 0xfb, 0xff, 0xff
};

// (x^3) * (35/2048383) + 127
static const rom unsigned char slow_drive_lut [] = {
    0x5c, 0x5d, 0x5e, 0x5e, 0x5f, 0x60, 0x61, 0x61, 
    0x62, 0x63, 0x64, 0x64, 0x65, 0x66, 0x66, 0x67, 
    0x68, 0x68, 0x69, 0x69, 0x6a, 0x6b, 0x6b, 0x6c, 
    0x6c, 0x6d, 0x6d, 0x6e, 0x6e, 0x6f, 0x6f, 0x70, 
    0x70, 0x71, 0x71, 0x72, 0x72, 0x73, 0x73, 0x73, 
    0x74, 0x74, 0x75, 0x75, 0x75, 0x76, 0x76, 0x76, 
    0x77, 0x77, 0x77, 0x77, 0x78, 0x78, 0x78, 0x79, 
    0x79, 0x79, 0x79, 0x7a, 0x7a, 0x7a, 0x7a, 0x7b, 
    0x7b, 0x7b, 0x7b, 0x7b, 0x7b, 0x7c, 0x7c, 0x7c, 
    0x7c, 0x7c, 0x7c, 0x7d, 0x7d, 0x7d, 0x7d, 0x7d, 
    0x7d, 0x7d, 0x7d, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 
    0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 0x7e, 
    0x7e, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x80, 0x80, 
    0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 
    0x80, 0x80, 0x80, 0x80, 0x81, 0x81, 0x81, 0x81, 
    0x81, 0x81, 0x81, 0x81, 0x82, 0x82, 0x82, 0x82, 
    0x82, 0x82, 0x83, 0x83, 0x83, 0x83, 0x83, 0x83, 
    0x84, 0x84, 0x84, 0x84, 0x85, 0x85, 0x85, 0x85, 
    0x86, 0x86, 0x86, 0x87, 0x87, 0x87, 0x87, 0x88, 
    0x88, 0x88, 0x89, 0x89, 0x89, 0x8a, 0x8a, 0x8b, 
    0x8b, 0x8b, 0x8c, 0x8c, 0x8d, 0x8d, 0x8e, 0x8e, 
    0x8f, 0x8f, 0x90, 0x90, 0x91, 0x91, 0x92, 0x92, 
    0x93, 0x93, 0x94, 0x95, 0x95, 0x96, 0x96, 0x97, 
    0x98, 0x98, 0x99, 0x9a, 0x9a, 0x9b, 0x9c, 0x9d, 
    0x9d, 0x9e, 0x9f, 0xa0, 0xa0, 0xa1, 0xa2, 0xa2
};

 	
  
     
  




/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
