/****************************
 * Cmucam3
 *
 * This is a modified version of the polly project with most features 
 * that cmucam2 has by custom writing
 * With this firmware loaded, the "GH" get histogram command will 
 * return a downsampled polly histogram.
 ****************************/

#include <cc3_math.h>
#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <cc3.h>
#include <cc3_ilp.h>
#include <cc3_img_writer.h>
#include <cc3_jpg.h>
#include <cc3_color_track.h>
#include <cc3_color_info.h>
#include <cc3_frame_diff.h>
#include "polly.h"
#include "packets.h"


#define BINS 28
static const unsigned int MAX_LINE = 128;
static const unsigned int MAX_ARGS = 10;

static const char VERSION_BANNER[] = "CMUcam3 v3.00 by jim bossert for FIRST team 1511";

//#define SERIAL_BAUD_RATE  CC3_UART_RATE_230400
#define SERIAL_BAUD_RATE  CC3_UART_RATE_115200
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_57600
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_38400
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_19200
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_9600
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_4800
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_2400
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_1200
//#define SERIAL_BAUD_RATE  CC3_UART_RATE_300



  bool raw_mode_output = false;
  bool raw_mode_no_confirmations = false;
  bool raw_mode_input = false;

typedef enum 
{
	BUFF_MODE,			//this command order the camera to continuely reload frame at 0 by fault, 1 one frame only
	READ_FRAME,			//this command order the camera to get new frame if BUFF_MODE disabled
	
	CAMERA_REG,			//
	CAMERA_POWER,		//
	CAMERA_TYPE,		//
	
	DELAY_MODE,			//this control the delay between data block over serial port
	POLL_MODE,			//this command continuely send packets at 1, 0 by default disabled, only send one
	PACKET_SKIP,		//
	RAW_MODE,			//send result in binary if enabled
	PACKET_FILTER,		//
	OUTPUT_MASK,		//
	
	SEND_FRAME,			//this command send a frame over serial port on F packet
	DOWN_SAMPLE,		//
	VIRTUAL_WINDOW,		//set the virtual window
	FRAME_STREAM,		//this command streaming frames over serial port, 0 by deault, disabled, 1 means abled
	HI_RES_MODE,		//this command enabled the hi-res frame on 1, by dault set at 0 disabled
	GET_WINDOW,			//this command get the setting of virtual window
	PIXEL_DIFF,			//
	
	TRACK_COLOR,		//this command order the camera to track color
	TRACK_INVERT,		//this causes tracking opposite of the defined color
	TRACK_WINDOW,		//this command track color in the virtual window
	NOISE_FILTER,		//
	LINE_MODE,			//
	GET_TRACK,			//this command get the color values used for tracking color
	SET_TRACK,			//set up settings for tracking color
	
	GET_HISTOGRAM,		//this command order the camera to send polly histogram back
	
	GET_VERSION,		//send the verison name back
	RESET,				//this command reset the camera
	
	SEND_JPEG,			//this command for send compressed jpeg back to viewer
	SET_MODE,			//this custom command set the modes of settings
	
	RETURN,                       // Must be second to last
	CMUCAM2_CMDS_COUNT            // Must be last entry so array sizes are correct
} cmucam2_command_t;

static const char cmucam2_cmds[CMUCAM2_CMDS_COUNT][3] = 
{
	[RETURN] = "",

	/* Buffer Commands */
	[BUFF_MODE] = "BM",
	[READ_FRAME] = "RF",
	
	/* Camera Module Commands */
	[CAMERA_REG] = "CR",
	[CAMERA_POWER] = "CP",
	[CAMERA_TYPE] = "CT",
	
	/* Data Rate Commands */
	[DELAY_MODE] = "DM",
	[POLL_MODE] = "PM",
	[PACKET_SKIP] = "PS",
	[RAW_MODE] = "RM",
	[PACKET_FILTER] = "PF",
	[OUTPUT_MASK] = "OM",
	
	/* Image Windowing Commands */
	[SEND_FRAME] = "SF",
	[DOWN_SAMPLE] = "DS",
	[VIRTUAL_WINDOW] = "VW",
	[FRAME_STREAM] = "FS",
	[HI_RES_MODE] = "HR",
	[GET_WINDOW] = "GW",
	[PIXEL_DIFF] = "PD",
	
	/* Auxiliary I/O Commands */
	
	/* Color Tracking Commands */
	[TRACK_COLOR] = "TC",
	[TRACK_INVERT] = "TI",
	[TRACK_WINDOW] = "TW",
	[NOISE_FILTER] = "NF",
	[LINE_MODE] = "LM",
	[GET_TRACK] = "GT",
	[SET_TRACK] = "ST",
	
	/* Histogram Commands */
	[GET_HISTOGRAM] = "GH",
	
	/* System Level Commands*/
	[GET_VERSION] = "GV",
	[RESET] = "RS",
	
	/* Custom commands */
	[SEND_JPEG] = "SJ",
	[SET_MODE] = "SM"
};

int32_t cmucam2_get_command (cmucam2_command_t * cmd, uint32_t * arg_list);

int main (void)
{
  uint32_t last_time, val,  step;
  bool change_settings, buff_mode, poll_mode, line_mode, packet_filter_flag = false;
  bool auto_led  = true;
  int settings_mode = 0;
  char c;
  uint8_t *x_axis, *h, cnt, conf;
  polly_config_t p_config;

  cc3_filesystem_init ();

  // configure uarts
  cc3_uart_init (0, SERIAL_BAUD_RATE, CC3_UART_MODE_8N1, CC3_UART_BINMODE_BINARY);
  // Make it so that stdout and stdin are not buffered
  val = setvbuf (stdout, NULL, _IONBF, 0);

  cc3_camera_init ();

  cc3_camera_set_power_state (true);
  cc3_camera_set_colorspace (CC3_COLORSPACE_RGB);
  cc3_camera_set_resolution (CC3_CAMERA_RESOLUTION_LOW);
  cc3_camera_set_auto_white_balance (true);
  cc3_camera_set_auto_exposure (true);


  cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, 2, 2);
  cc3_pixbuf_frame_set_coi (CC3_CHANNEL_GREEN);

  cc3_led_set_state (0, false);
  cc3_led_set_state (1, false);
  cc3_led_set_state (2, false);
  
  raw_mode_output = false;
  raw_mode_no_confirmations = false;
  raw_mode_input = false;

  // sample wait command in ms 
  cc3_timer_wait_ms (1000);

  // initialize pixbuf
  cc3_pixbuf_load ();

  x_axis = malloc (cc3_g_pixbuf_frame.width);
  h = malloc (cc3_g_pixbuf_frame.width);

  p_config.color_thresh = 10;
  p_config.min_blob_size = 20;
  p_config.connectivity = 1;
  p_config.horizontal_edges = 1;
  p_config.vertical_edges = 1;
  p_config.blur = 1;
  p_config.histogram = malloc (cc3_g_pixbuf_frame.width);

  printf ("%s", VERSION_BANNER);
  	while (1) 
  		{
    		uint32_t val, n, step, go;
    		cmucam2_command_t command;
    		uint32_t arg_list[MAX_ARGS];
    		printf (":");
    		// Parse out the GH command using the CMUcam2 command parser
    		n = cmucam2_get_command (&command, arg_list);
    			if (n != -1) 
				{
					if (change_settings == true)
					{
						if (settings_mode == 0)//settings for polly
						{
							cc3_camera_set_colorspace (CC3_COLORSPACE_RGB);
							cc3_camera_set_resolution (CC3_CAMERA_RESOLUTION_LOW);
							cc3_camera_set_auto_white_balance (true);
							cc3_camera_set_auto_exposure (true);
							
							cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, 2, 2);
							cc3_pixbuf_frame_set_coi (CC3_CHANNEL_GREEN);
							printf ("ACK\r");
						}
						else if (settings_mode == 1)// settings for color tracking in RGB
						{
							cc3_camera_set_resolution (CC3_CAMERA_RESOLUTION_LOW);
		      				cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, 2, 1);
		      				
							printf ("ACK\r");
						}
						else if (settings_mode == 2)
						{
							
						}
						else 
						{
							printf ("NCK\r");
						}
						
					change_settings = false;
					}
      				// If it is the get histogram command, then run Polly Algorithm
      				if (command == GET_HISTOGRAM) 
      				{
        				printf ("ACK\r");
        				go = true;
        				step = cc3_g_pixbuf_frame.width / BINS;
        				step--;
						// Continue to load histograms until a '\r' is detected
        					while (go) 
        					{
	  							// Load a frame
          						cc3_pixbuf_load ();
	  							// Run polly on the frame
          						polly (p_config);
          						// p_config.histogram gets filled with the return data
          						printf ("H");
          							for (int i = 0; i < BINS; i++) 
          							{
            							// Downsample bins to 28 so that it is viewable in the CMUcam2 GUI
	    								// Use MIN of multiple bins
            							val = 255;
            							for (uint32_t j = 0; j < step; j++)
              							if (p_config.histogram[(i * step) + j] < val)
                						val = p_config.histogram[(i * step) + j];
            							printf (" %d", val);
          							}	
          						printf ("\r");
          						// Check for data in Uart
          							while (!cc3_uart_has_data (0)) 
          							{
            							if (getchar () == '\r') 
            							{
              								go = false;
              								break;
            							}
          							}
        					}
      				}
      				else if (command == SEND_JPEG)
      				{
	      				printf("ACK\r");
	      				cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, 1, 1);
	      				cc3_pixbuf_frame_set_coi (CC3_CHANNEL_ALL);
	      				cc3_timer_wait_ms (500);
	      				cc3_jpeg_send_simple ();
	      				printf("JPG_END\r");
      				}
      				else if (command == BUFF_MODE)
      				{
	      				printf ("ACK\r");
	      				if (arg_list[0] == 1)
	      					buff_mode = true;
	      				else
	      					buff_mode = false;
      				}
      				else if (command == SET_MODE)
      				{
	      				if (arg_list[0] == 0)
	      				{
		      				settings_mode = 0;
		      				change_settings = true;
		      				printf ("ACK\r");
	      				}
	      				else if (arg_list[0] == 1)
	      				{
		      				settings_mode = 1;
		      				change_settings = true;
		      				printf ("ACK\r");
	      				}
	      				else if (arg_list[0] == 2)
	      				{
		      				settings_mode = 2;
		      				change_settings = true;
		      				printf ("ACK\r");
	      				}
	      				else if (arg_list[0] == ' ')
	      				{
		      				printf ("NCK\r");
	      				}
	      				else 
	      				{
		      				printf ("NCK\r");
	      				}
      				}
      				else if (command == GET_WINDOW)
      				{
	      				printf ("ACK\r");
	      				if (raw_mode_output) 
	      				{
		      				putchar (255);
		      				raw_print(cc3_g_pixbuf_frame.x0 / 2);
		      				raw_print(cc3_g_pixbuf_frame.y0);
		      				raw_print(cc3_g_pixbuf_frame.x1 / 2);
		      				raw_print(cc3_g_pixbuf_frame.y1);
		      			}
		      			else 
		      			{
			      			printf ("%d %d %d %d\r", cc3_g_pixbuf_frame.x0 / 2, cc3_g_pixbuf_frame.y0, cc3_g_pixbuf_frame.x1 / 2, cc3_g_pixbuf_frame.y1);
		      			}
	      			}
	      			else if (command == CAMERA_POWER)
	      			{
		      			printf ("ACK\r");
		      			{
		      			// save
		      			uint16_t x_0 = cc3_g_pixbuf_frame.x0;
		      			uint16_t y_0 = cc3_g_pixbuf_frame.y0;
		      			uint16_t x_1 = cc3_g_pixbuf_frame.x1;
		      			uint16_t y_1 = cc3_g_pixbuf_frame.y1;
		      			uint8_t x_step = cc3_g_pixbuf_frame.x_step;
		      			uint8_t y_step = cc3_g_pixbuf_frame.y_step;
		      			cc3_subsample_mode_t ss_mode = cc3_g_pixbuf_frame.subsample_mode;
		      			cc3_camera_set_power_state (arg_list[0]);

		      			// restore
		      			cc3_pixbuf_frame_set_roi (x_0, y_0, x_1, y_1);
		      			cc3_pixbuf_frame_set_subsample (ss_mode, x_step, y_step);
	      				}
	      			}
      				else if (command == DOWN_SAMPLE)
      				{
	      				printf ("ACK\r");
	      				cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, arg_list[0] * 2, arg_list[1]);
      				}
      				else if (command == CAMERA_REG)
      				{
	      				printf ("ACK\r");
	      				for (int i = 0; i < n; i += 2)
	      					cc3_camera_set_raw_register (arg_list[i], arg_list[i + 1]);
      				}
      				else if (command == PACKET_FILTER)
      				{
	      				printf ("ACK\r");
	      				if (arg_list[0] == 1)
	      					packet_filter_flag = true;
	      				else
	      					packet_filter_flag = false;
      				}
      				else if (command == POLL_MODE)
      				{
	      				printf ("ACK\r");
	      				if (arg_list[0] == 1)
	      					poll_mode = true;
	      				else
	      					poll_mode = false;
      				}
      				else if (command == HI_RES_MODE)
      				{
	      				printf ("ACK\r");
	      				if (arg_list[0] == 1)
	      				{
		      				cc3_camera_set_resolution (CC3_CAMERA_RESOLUTION_HIGH);
	      				}
	      				else if (arg_list[0] == 0)
	      				{
		      				cc3_camera_set_resolution (CC3_CAMERA_RESOLUTION_LOW);
		      				cc3_pixbuf_frame_set_subsample (CC3_SUBSAMPLE_NEAREST, 2, 1);
	      				}
	      				else 
	      				{
		      				printf ("NCK\r");
	      				}
      				}
      				
      				else if (command == RETURN)
      				{
      					printf ("ACK\r");
  					}
      				else if (command == GET_VERSION)
      				{
        				printf ("ACK\r");
        				printf("%s", VERSION_BANNER);
    				}
      				else if (command == RESET)
      				{
        				printf ("ACK\r");
        				printf ("%s", VERSION_BANNER);
    				}
      				else
        				printf ("NCK\r");
    			}
    			else
      				printf ("NCK\r");
  }

  return 0;
}


// CMUcam2 command parser
int32_t cmucam2_get_command (cmucam2_command_t * cmd, uint32_t * arg_list)
{
  char line_buf[MAX_LINE];
  int c;
  char *token;
  bool fail = false;
  uint32_t length, argc;
  uint32_t i;

  length = 0;
  c = 0;
  while (c != '\r') 
  	{
    	c = getchar ();

    	if (length < (MAX_LINE - 1) && c != EOF) 
    	{
      		line_buf[length] = c;
      		length++;
    	}
    	else 
    	{
      		// too long or EOF
      		return -1;
    	}
  	}
  // null terminate
  line_buf[length] = '\0';

  // check for empty command
  if (line_buf[0] == '\r' || line_buf[0] == '\n') 
  {
    *cmd = RETURN;
    return 0;
  }

  // start looking for command
  token = strtok (line_buf, " \r\n");

  if (token == NULL) 
  {
    // no command ?
    return -1;
  }

  // get name of the command
  for (i = 0; i < strlen (token); i++) 
  {
    token[i] = toupper (token[i]);
  }

  // do lookup of command
  fail = true;
  for (i = 0; i < CMUCAM2_CMDS_COUNT; i++) 
  {
    if (strcmp (token, cmucam2_cmds[i]) == 0) 
    {
      fail = false;
      *cmd = i;
      break;
    }
  }
  if (fail) 
  {
    return -1;
  }

  // now get the arguments
  argc = 0;
  while (true) 
  {
    // extract string from string sequence
    token = strtok (NULL, " \r\n");
    // check if there is nothing else to extract
    	if (token == NULL) 
    	{
      		// printf("Tokenizing complete\n");
      		return argc;
    	}

    // make sure the argument is fully numeric
    	for (i = 0; i < strlen (token); i++) 
    		{
      			if (!isdigit (token[i]))
        		return -1;
    		}

    // we have a valid token, add it
    arg_list[argc] = atoi (token);
    argc++;
  }

  return -1;
}
