/*******************************************************************************
*
*	TITLE:		control.c 
*
*	VERSION:	0.1                           
*
*	DATE:		01-29-2008
*
*	AUTHOR:		Team 1511
*
*	DESCRIPTION:	This file contains functions for controlling the arm and manipulator.
*******************************************************************************/

#include <stdio.h>
#include "ifi_aliases.h"
#include "ifi_default.h"
#include "ifi_utilities.h"
#include "control.h"
#include "iomap.h"
#include "motion.h"
#include "user_routines.h"

#define TRAVERSE_UP 1
#define TRAVERSE_DOWN 2

static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm);

static int Arm_Process(int arm, int wtarget, int gripper);
static unsigned char Wrist_PID(unsigned int target);
static int Raise_Arm(void);
static int Lower_Arm(void);
static void Open_Manip(void);
static void Close_Manip(void);
static void Set_Pump_Precharge(int ticks);

unsigned int wrist_target = WRIST_POT_PARK;
static PIDData WristPID;
static int pump_precharge_ticks = 0;
static unsigned char arm_prev_up = 0;
static unsigned char manip_prev_open = 0;
static int arm_traverse = 0;

void Initialize_Arm(void)
{
    wrist_target = WRIST_POT_PARK;
    Init_Wrist_PIDData_Struct(&WristPID, &pwm_wrist);

    arm_prev_up = !oi_broken_shoulder &&
           (anain_shoulder_pot > (ARM_POT_MAX - ARM_POT_THRESHOLD));
//    manip_prev_open = !digin_gripper_open;
    manip_prev_open = 0;
}

static void Init_Wrist_PIDData_Struct(PIDData* control, unsigned char* pwm)
{
  // p = want - am;
  // i = sum(want - am over time)
  // output = kp * p + ki * i;

  Init_PIDData_Struct(control, pwm);
  control->tolerance = 14;
//  control->kp_num = -10;
  control->kp_num = -15;
  control->kp_den = 20;
  control->ki_num = -3;
  control->ki_den = 200;
  control->kd_num = 0;
  control->kd_den = 1;
  control->int_err_max = 10000;
  control->p_out_max = 80;
  control->i_out_max = 45;
//  control->output_min = 30;
  control->output_min = 32;
//  control->output_max = 90;
  control->output_max = 127;
}

int AutoArm(int target)
{
    int arm, wtarget, gripper;
    if ( target == AUTOARM_TARGET_PARK)
    {
        arm = 2;
        wtarget = WRIST_POT_PARK;
        gripper = 2;
    }
    else if (target == AUTOARM_TARGET_PICKUP)
    {
        arm = 2;
        wtarget = WRIST_POT_PICKUP;
        gripper = 1;
    }
    else if (target == AUTOARM_TARGET_HURDLE)
    {
        arm = 1;
        wtarget = WRIST_POT_HURDLE;
        gripper = 0;
    }
	else if (target == AUTOARM_TARGET_HURDLE_FINISH)
	{
		arm = 1;
		wtarget = WRIST_POT_HURDLE_FINISH;
		if (!oi_broken_wrist && anain_wrist_pot < WRIST_POT_HURDLE_OPEN)
			gripper = 1;
		else
			gripper = 0;
	}
    else if (target == AUTOARM_TARGET_HITBALL)
    {
        arm = 1;
        wtarget = WRIST_POT_MAX;
        gripper = 2;
    }
    else if (target == AUTOARM_TARGET_PICK)
    {
        arm = 1;
        wtarget = WRIST_POT_MIN;
        gripper = 1;
    }
    else
    {
        arm = 0;
        wtarget = wrist_target;
        gripper = 0;
    }

    return Arm_Process(arm, wtarget, gripper);
}

// arm == 0 -- keep stable
// arm == 1 -- arm up
// arm == 2 -- arm down
// wtarget == positive, go to target - ignored if broken wrist
// gripper == 0 keep stable
// gripper == 1 open
// gripper == 2 close
int Arm_Process(int arm, int wtarget, int gripper)
{
	int arm_done = 0;
	int wrist_done = 0;

    if (arm == 1)
        Raise_Arm();
    else if (arm == 2)
        Lower_Arm();
    else
    {
        // Continue to raise/lower arm even if controls not held.
        if (arm_traverse == TRAVERSE_UP)
            Raise_Arm();
        else if (arm_traverse == TRAVERSE_DOWN)
            Lower_Arm();
    }

    // Check for completed arm traversal.
    // arm_traverse only set if !oi_broken_shoulder, so
    // no need to check it here.
    if (arm_traverse == TRAVERSE_UP && 
		anain_shoulder_pot > (ARM_POT_MAX - ARM_POT_THRESHOLD))
	{
        arm_traverse = 0;
		arm_done = 1;
    }
	else if (arm_traverse == TRAVERSE_DOWN &&
        	 anain_shoulder_pot < (ARM_POT_MIN + ARM_POT_THRESHOLD))
	{
    	arm_traverse = 0;
		arm_done = 1;
    }
	else if (!arm_traverse)
	{
		arm_done = 1;
	}
    
    if (gripper == 1)
        Open_Manip();
    else if (gripper == 2)
        Close_Manip();

    if (oi_broken_wrist)
    {
        if (oi_wrist_up)
            pwm_wrist = 127 - WRIST_MANUAL_SPEED;
        else if (oi_wrist_down)
            pwm_wrist = 127 + WRIST_MANUAL_SPEED;
        else
            pwm_wrist = 127;

		wrist_done = 1;
    }
    else
    {
        wrist_target = wtarget;
        
        if (wrist_target < WRIST_POT_MIN && arm_prev_up)
            wrist_target = WRIST_POT_MIN;
        else if (wrist_target < WRIST_POT_MIN_FLOOR && !arm_prev_up)
            wrist_target = WRIST_POT_MIN_FLOOR;
        else if (wrist_target > WRIST_POT_MAX)
            wrist_target = WRIST_POT_MAX;
        else if ((arm_traverse == TRAVERSE_UP) && (wrist_target < WRIST_POT_MIN_RAISE))
            wrist_target = WRIST_POT_MIN_RAISE;
         else if ((arm_traverse == TRAVERSE_DOWN) && (wrist_target < WRIST_POT_MIN_LOWER))
            wrist_target = WRIST_POT_MIN_LOWER;       
    
        wrist_done = Wrist_PID(wrist_target);
    }

    if (pump_precharge_ticks)
        --pump_precharge_ticks;

	return arm_done && wrist_done;
}

void Arm_Control(void)
{
	int arm, wtarget, gripper;

    if (oi_arm > 127 + ARM_CONTROL_DEADBAND)
        arm = 2;
    else if (oi_arm < 127 - ARM_CONTROL_DEADBAND)
        arm = 1;
    else
        arm = 0;

    if (oi_gripper_open)
        gripper = 1;
    else if (oi_gripper_close)
        gripper = 2;
	else
		gripper = 0;

    if (oi_wrist_up)
     wtarget = (wrist_target + WRIST_TARGET_STEP);
    else if (oi_wrist_down)
     wtarget = (wrist_target - WRIST_TARGET_STEP);
	else
	 wtarget = wrist_target;

	Arm_Process(arm, wtarget, gripper);

}

static int Raise_Arm(void)
{
    if (!oi_broken_wrist && !oi_broken_shoulder)
        arm_traverse = TRAVERSE_UP;

    if (anain_wrist_pot < WRIST_POT_MIN_RAISE && 
        !oi_broken_wrist && !oi_broken_shoulder)
    {
        if (wrist_target < WRIST_POT_MIN_RAISE)
            wrist_target = WRIST_POT_MIN_RAISE;

    }
    else
    {
        relay_arm_up = 1;
        relay_arm_down = 0;
        if (!arm_prev_up)
        {
            Set_Pump_Precharge(ARM_PUMP_TICKS);
            arm_prev_up = 1;
        }
    }
    // XXX - remove ret val
    return 1;
}

static int Lower_Arm(void)
{
    if (!oi_broken_wrist && !oi_broken_shoulder)
        arm_traverse = TRAVERSE_DOWN;

    // XXX - Raise and lower copied code
    if (anain_wrist_pot < WRIST_POT_MIN_RAISE && 
        !oi_broken_wrist && !oi_broken_shoulder)
    {
        if (wrist_target < WRIST_POT_MIN_RAISE)
            wrist_target = WRIST_POT_MIN_RAISE;

    }
    else
    {
        relay_arm_up = 0;
        relay_arm_down = 1;
        arm_prev_up = 0;
    }

    // XXX - remove ret val
    return 1;
}

static unsigned char Wrist_PID(unsigned int target)
{
    static unsigned int previous_position = 0;
    unsigned char ret;

    if (previous_position != target)
        Set_PIDData_Target(&WristPID, target);

    previous_position = target;

    ret = PID_Control(&WristPID, anain_wrist_pot);

    return ret;
}

static void Open_Manip(void)
{
    relay_gripper_open = 1;
    relay_gripper_close = 0;
    if (!manip_prev_open)
    {
        Set_Pump_Precharge(MANIP_PUMP_TICKS);
        manip_prev_open = 1;
    }
}

static void Close_Manip(void)
{
    relay_gripper_open = 0;
    relay_gripper_close = 1;
    if (manip_prev_open)
    {
        Set_Pump_Precharge(MANIP_PUMP_TICKS);
        manip_prev_open = 0;
    }
}

static void Set_Pump_Precharge(int ticks)
{
    if (pump_precharge_ticks < ticks)
        pump_precharge_ticks = ticks;
}

unsigned char Pump_Precharge(void)
{
    return (unsigned char)(pump_precharge_ticks > 0);
}
